require "Apollo"
require "ApolloTimer"

local AlpacaPluginLibrary = {}

function AlpacaPluginLibrary:new(o)
  o = o or {}
  setmetatable(o, self)
  self.__index = self 
  return o
end

function AlpacaPluginLibrary:Init()
  self.arPending = {}
  self.arRegistered = {}
  
  Apollo.RegisterPackage(self, "AlpacaPluginLibrary", 1, {})
end

function AlpacaPluginLibrary:OnLoad() 
  Apollo.RegisterEventHandler("AlpacaRegisterPluginSuccess", "OnAlpacaRegisterPluginSuccess", self)
end

function AlpacaPluginLibrary:ToggleTimer()
  if #self.arPending == 0 then
    self.timer = nil
  elseif not self.timer then
    self.timer = ApolloTimer.Create(0.01, true, "OnTimer", self)
  end  
end

function AlpacaPluginLibrary:OnTimer()  
  for _, objPending in ipairs(self.arPending) do
    Event_FireGenericEvent("AlpacaRegisterPlugin", objPending)     
  end
end

function AlpacaPluginLibrary:RegisterPlugin(objPlugin)
  table.insert(self.arPending, objPlugin) 
  self:ToggleTimer()  
end

function AlpacaPluginLibrary:OnAlpacaRegisterPluginSuccess(objPlugin)
  local bIsPending = false
  
  for i, objPending in ipairs(self.arPending) do
    if objPending == objPlugin then
      table.insert(self.arRegistered, table.remove(self.arPending, i))
      self:ToggleTimer()
    end
  end
end

-- Debug

function AlpacaPluginLibrary:SetDebugCounter(strName, fValue)
  Event_FireGenericEvent("AlpacaSetDebugCounter", strName, fValue)
end













AlpacaPluginLibrary.EnumInputType = {
  Header = 1,
  SubHeader = 2,
  Checkbox = 3,
  Color = 4,
  Select = 5,
  Slider = 6
}


-- Form

function AlpacaPluginLibrary:CreateForm(arInputDefinitions, arSubForms)
  return {
    arInputDefinitions = arInputDefinitions,
    arSubForms = arSubForms   
  }
end

function AlpacaPluginLibrary:CreateSubForm(strLabel, arInputDefinitions)
  return {
    strLabel = strLabel,
    arInputDefinitions = arInputDefinitions
  }
end

function AlpacaPluginLibrary:CreateHeader(strLabel) 
  return {
    eInputType = self.EnumInputType.Header,
    strLabel = strLabel
  }
end

function AlpacaPluginLibrary:CreateSubHeader(strLabel) 
  return {
    eInputType = self.EnumInputType.SubHeader,
    strLabel = strLabel
  }
end

function AlpacaPluginLibrary:CreateCheckbox(strLabel, arSettingPath)
  return {
    eInputType = self.EnumInputType.Checkbox,
    strLabel = strLabel,
    arSettingPath = arSettingPath
  }  
end

function AlpacaPluginLibrary:CreateColor(strLabel, arSettingPath)
  return {
    eInputType = self.EnumInputType.Color,
    strLabel = strLabel,
    arSettingPath = arSettingPath
  }  
end

function AlpacaPluginLibrary:CreateSelect(strLabel, arSettingPath, arOptions)
  return {
    eInputType = self.EnumInputType.Select,
    strLabel = strLabel,
    arSettingPath = arSettingPath,
    arOptions = arOptions
  }
end

function AlpacaPluginLibrary:CreateSlider(strLabel, arSettingPath, fMin, fMax, fStep)
  return {
    eInputType = self.EnumInputType.Slider,
    strLabel = strLabel,
    arSettingPath = arSettingPath,
    fMin = fMin,
    fMax = fMax,
    fStep = fStep
  }
end

-- Start

local AlpacaPluginLibraryInst = AlpacaPluginLibrary:new()
AlpacaPluginLibraryInst:Init()
