require "Apollo"
require "GameLib"

local AlpacaGameLibrary = nil
local AlpacaPluginLibrary = nil
local AlpacaStandardLibrary = nil
local AlpacaWindowLibrary = nil

local AlpacaActions = {} 

local knMaxVisibleOptions = 5

local keNonCombatActionType = {
  Innate = 1,
  Potion = 2,
  Path = 3,
  Mount = 4,
  Recall = 5
}

-- Plugin

function AlpacaActions:PluginId()
  return "AlpacaActions"
end

function AlpacaActions:PluginName()
  return "Actions"
end

function AlpacaActions:PluginDefaults() 
  local fnCreateColor = function(...)
    return AlpacaWindowLibrary:AlpacaColor(...)
  end
 
  local tSettings = {
    bEnable = true,
    tCombatActionBar = {
      bOnlyShowInCombat = false,
      nCutoff = 10,
      fCooldownAlpha = 0.5,
      tPosition = {
        nHorizontal = 500,
        nVertical = 660
      },
      nSpacing = 30,
      nIconSize = 30,
      tBarSize = {
        nWidth = 10,
        nHeight = 80
      },
      crBarColor = AlpacaWindowLibrary:AlpacaColor(1, 1, 1, 0.3)
    },
    tNonCombatActionBar = {
      bOnlyShowInCombat = false,
      tPosition = {
          nHorizontal = 500,
          nVertical = 800
      },
      nSpacing = 5,
      nIconSize = 30,
      bShowInnate = true,
      bShowPotion = true,
      bShowPath = true,
      bShowRecall = true,
      bShowMount = true
    }
  }  
    
  return tSettings
end

function AlpacaActions:PluginForm() 
  local arInputDefinitions = {
    AlpacaPluginLibrary:CreateHeader("Actions"),
    AlpacaPluginLibrary:CreateCheckbox("Enable",                { "bEnable" }),
    
    AlpacaPluginLibrary:CreateSubHeader("Combat Bar"),
    AlpacaPluginLibrary:CreateCheckbox("Only Show In Combat",   { "tCombatActionBar", "bOnlyShowInCombat" }),
    AlpacaPluginLibrary:CreateSlider("Position (Horizontal)",   { "tCombatActionBar", "tPosition", "nHorizontal" }, 0, 2000, 1),
    AlpacaPluginLibrary:CreateSlider("Position (Vertical)",     { "tCombatActionBar", "tPosition", "nVertical" }, 0, 2000, 1),
    AlpacaPluginLibrary:CreateSlider("Icon Size",               { "tCombatActionBar", "nIconSize" }, 10, 200, 1),
    AlpacaPluginLibrary:CreateSlider("Spacing",                 { "tCombatActionBar", "nSpacing" }, 0, 200, 1),
    AlpacaPluginLibrary:CreateSlider("Bar Size (Width)",        { "tCombatActionBar", "tBarSize", "nWidth" }, 0, 100, 1),
    AlpacaPluginLibrary:CreateSlider("Bar Size (Height)",       { "tCombatActionBar", "tBarSize", "nHeight" }, 0, 500, 1),
    AlpacaPluginLibrary:CreateColor("Bar Color",                { "tCombatActionBar", "crBarColor" }),
    AlpacaPluginLibrary:CreateSlider("Cooldown Threshold",      { "tCombatActionBar", "nCutoff" }, 0, 120, 1),
    AlpacaPluginLibrary:CreateSlider("Cooldown Alpha",          { "tCombatActionBar", "fCooldownAlpha" }, 0, 1, 0.01),
    
    AlpacaPluginLibrary:CreateSubHeader("Non-Combat Bar"),
    AlpacaPluginLibrary:CreateCheckbox("Only Show In Combat",   { "tNonCombatActionBar", "bOnlyShowInCombat" }),
    AlpacaPluginLibrary:CreateCheckbox("Innate",                { "tNonCombatActionBar", "bShowInnate" }),
    AlpacaPluginLibrary:CreateCheckbox("Potion",                { "tNonCombatActionBar", "bShowPotion" }),
    AlpacaPluginLibrary:CreateCheckbox("Path",                  { "tNonCombatActionBar", "bShowPath" }),
    AlpacaPluginLibrary:CreateCheckbox("Recall",                { "tNonCombatActionBar", "bShowRecall" }),
    AlpacaPluginLibrary:CreateCheckbox("Mount",                 { "tNonCombatActionBar", "bShowMount" }),
    AlpacaPluginLibrary:CreateSlider("Position (Horizontal)",   { "tNonCombatActionBar", "tPosition", "nHorizontal" }, 0, 2000, 1),
    AlpacaPluginLibrary:CreateSlider("Position (Vertical)",     { "tNonCombatActionBar", "tPosition", "nVertical" }, 0, 2000, 1),
    AlpacaPluginLibrary:CreateSlider("Icon Size",               { "tNonCombatActionBar", "nIconSize" }, 10, 200, 1),
    AlpacaPluginLibrary:CreateSlider("Spacing",                 { "tNonCombatActionBar", "nSpacing" }, 0, 200, 1),
    
  }
  
  return AlpacaPluginLibrary:CreateForm(arInputDefinitions)
end

function AlpacaActions:OnPluginRegistered(wndDisplay) 
  self.wndDisplay = wndDisplay 
end

function AlpacaActions:OnPluginSettingsChanged(tSettings)
  self.tSettings = tSettings
  self:RequestRebuild(0)
end

-- Addon

function AlpacaActions:new(o)
  o = o or {}
  setmetatable(o, self)
  self.__index = self 
  return o
end

function AlpacaActions:Init() 
  local bHasConfigureFunction = false
  local strConfigureButtonText = nil
  local tDependencies = { "AlpacaGameLibrary", "AlpacaPluginLibrary", "AlpacaStandardLibrary", "AlpacaWindowLibrary" }
  
  Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end
 
function AlpacaActions:OnLoad()
  Apollo.LoadSprites("AlpacaSprites.xml")

  AlpacaGameLibrary = Apollo.GetPackage("AlpacaGameLibrary").tPackage
  AlpacaPluginLibrary = Apollo.GetPackage("AlpacaPluginLibrary").tPackage
  AlpacaStandardLibrary = Apollo.GetPackage("AlpacaStandardLibrary").tPackage
  AlpacaWindowLibrary = Apollo.GetPackage("AlpacaWindowLibrary").tPackage
  
  Apollo.RegisterEventHandler("VarChange_FrameCount", "OnFrame", self)
  
  Apollo.RegisterEventHandler("PlayerChanged", "OnPlayerChanged", self)
  Apollo.RegisterEventHandler("CharacterCreated", "OnCharacterCreated", self)
  Apollo.RegisterEventHandler("PlayerLevelChange", "OnPlayerLevelChange", self)
  Apollo.RegisterEventHandler("AbilityBookChange", "OnAbilityBookChange", self)
  Apollo.RegisterEventHandler("SpecChanged", "OnSpecChanged", self)
  
  self.xmlDoc = XmlDoc.CreateFromFile("AlpacaActions.xml")
  
  AlpacaPluginLibrary:RegisterPlugin(self, "Actions")
end

function AlpacaActions:OnPlayerChanged()
  self:RequestRebuild(1)
end

function AlpacaActions:OnCharacterCreated()
  self:RequestRebuild(1)
end

function AlpacaActions:OnPlayerLevelChange()
  self:RequestRebuild(1)
end

function AlpacaActions:OnAbilityBookChange()
  self:RequestRebuild(1)
end

function AlpacaActions:OnSpecChanged()
  self:RequestRebuild(1)
end

-- Timers

function AlpacaActions:RequestRebuild(fDelay)
  self.fRebuildTime = os.clock() + (fDelay or 0)
end

function AlpacaActions:IsReady() 
  if not self.bReady then
    if not AlpacaWindowLibrary:IsLoaded() then 
      return false
    end
      
    if not self.xmlDoc then 
      return false
    end
    
    if not self.wndDisplay then
      return false
    end
    
    if not self.tSettings then
      return false
    end

    self.bReady = true
  end
  
  return self.bReady
end

function AlpacaActions:OnFrame()  
  if not self:IsReady() then
    return
  end
  
  self.unitPlayer = AlpacaGameLibrary:GetPlayerUnit()

  if not self.unitPlayer then
    return
  end

  local bRebuild = self.fRebuildTime and os.clock() >= self.fRebuildTime 

  if not self.wndCombatActionBar or bRebuild then
    self:CreateCombatActionBar()
  end

  if not self.wndNonCombatActionBar or bRebuild then
    self:CreateNonCombatActionBar()
  end
    
  if bRebuild then
    self.fRebuildTime = nil
  end
  
  self:RefreshCombatActionBar()
  self:RefreshNonCombatActionBar()
end

-- CombatActionBar

function AlpacaActions:CreateCombatActionBar() 
  if self.wndCombatActionBar then
    self.wndCombatActionBar:Destroy()
    self.wndCombatActionBar = nil
  end
  
  if not self.tSettings.bEnable then
    return
  end
  
  local wndCombatActionBar = AlpacaWindowLibrary:CreateEmptyFrame(self.wndDisplay)  
     
  self:CreateCombatAction(wndCombatActionBar, "InnateButton")

  for i = 0, 8 do
      if ActionSetLib.IsSlotUnlocked(i) == ActionSetLib.CodeEnumLimitedActionSetResult.Ok then        
        self:CreateCombatAction(wndCombatActionBar, "LASButton", i)
      end
  end
  
  local nHorizontal = self.tSettings.tCombatActionBar.tPosition.nHorizontal
  local nVertical = self.tSettings.tCombatActionBar.tPosition.nVertical
  local nSpacing = self.tSettings.tCombatActionBar.nSpacing  
  
  self:ResizeContainer(wndCombatActionBar, nHorizontal, nVertical, nSpacing)
  
  self.wndCombatActionBar = wndCombatActionBar
end

function AlpacaActions:CreateCombatAction(wndContainer, strButton, nContentId)
  local nIconSize = self.tSettings.tCombatActionBar.nIconSize
  local nBarWidth = self.tSettings.tCombatActionBar.tBarSize.nWidth
  local nBarHeight = self.tSettings.tCombatActionBar.tBarSize.nHeight
  local crBarColor = self.tSettings.tCombatActionBar.crBarColor
  
  local wndAction = Apollo.LoadForm(self.xmlDoc, "CombatAction", wndContainer, self)
    
  if strButton ~= "InnateButton" then
    wndAction:FindChild("InnateButton"):Destroy()
  else
    wndAction:FindChild("InnateButton"):SetName("Button")
  end
  
  if strButton ~= "LASButton" then
    wndAction:FindChild("LASButton"):Destroy()    
  else
    wndAction:FindChild("LASButton"):SetName("Button")
    wndAction:FindChild("Button"):SetContentId(nContentId)    
  end
  
  local nButtonLeft = 0 - math.floor(nIconSize / 2)
  local nButtonTop = 0
  local nButtonRight = nButtonLeft + nIconSize
  local nButtonBottom = nButtonTop + nIconSize
  
  wndAction:SetAnchorPoints(0, 0, 0, 0)
  wndAction:SetAnchorOffsets(0, 0, math.max(nBarWidth, nIconSize), math.max(nBarHeight, nIconSize))
  wndAction:FindChild("Button"):SetAnchorOffsets(nButtonLeft, nButtonTop, nButtonRight, nButtonBottom)
  
  local wndBar = wndAction:FindChild("Bar")
  
  wndBar:SetBGColor(AlpacaWindowLibrary:ToApolloColor(crBarColor))
  wndBar:SetAnchorOffsets(-1 * math.floor(nBarWidth / 2), 0, math.ceil(nBarWidth / 2), 0)
end

function AlpacaActions:RefreshCombatActionBar()
  if not self.wndCombatActionBar then
    return
  end
  
  local bShow = true
  
  if self.tSettings.tCombatActionBar.bOnlyShowInCombat and not AlpacaGameLibrary:IsInCombatOrDamaged(self.unitPlayer) then
    bShow = false
  end
  
  if self.tSettings.bPreview then
    bShow = true
  end
  
  local wndCombatActionBar = self.wndCombatActionBar

  wndCombatActionBar:Show(bShow, true)
  
  if bShow then
    for i, wndAction in ipairs(wndCombatActionBar:GetChildren()) do    
      self:RefreshCombatAction(wndAction)
    end
  end
end

function AlpacaActions:RefreshCombatAction(wndAction)
  local objSpell = wndAction:FindChild("Button"):GetContent().spell
 
  if not objSpell then
    return
  end
  
  local fCooldownRemaining = objSpell:GetCooldownRemaining()
  local tAbilityCharges = objSpell:GetAbilityCharges()
  local nChargesMax = 0	
  local nChargesRemaining = 0
  local fRechargeRemaining = 0
  if (tAbilityCharges) then
	nChargesMax = tAbilityCharges.nChargesMax
	nChargesRemaining = tAbilityCharges.nChargesRemaining
	fRechargeRemaining = tAbilityCharges.fRechargeTime * tAbilityCharges.fRechargePercentRemaining
  end
         
  if nChargesMax > 0 then
    if nChargesRemaining == 0 then
      self:MoveButton(wndAction:FindChild("Button"), math.min(1, fRechargeRemaining / self.tSettings.tCombatActionBar.nCutoff))
    else
      self:MoveButton(wndAction:FindChild("Button"), 0)
    end
  else
    self:MoveButton(wndAction:FindChild("Button"), math.min(1, fCooldownRemaining / self.tSettings.tCombatActionBar.nCutoff))
  end
end

function AlpacaActions:MoveButton(wndButton, fPercent)
  local wndRange = wndButton:GetParent()
  
  local nRangeHeight = wndRange:GetHeight()
  
  local nIconWidth = wndButton:GetWidth()
  local nIconHeight = wndButton:GetHeight()
  
  local nLeft = -1 * math.ceil(nIconWidth / 2)
  local nTop = math.floor((fPercent) * (nRangeHeight - nIconHeight))
  local nRight = nLeft + nIconWidth
  local nBottom = nTop + nIconHeight
  
  wndButton:SetAnchorOffsets(nLeft, nTop, nRight, nBottom) 
  
  if fPercent == 0 then
    wndButton:SetOpacity(1)
  else
    wndButton:SetOpacity(self.tSettings.fCooldownAlpha)
  end
end

-- NonCombatActionBar

function AlpacaActions:CreateNonCombatActionBar()
  if self.wndNonCombatActionBar then
    self.wndNonCombatActionBar:Destroy()
    self.wndNonCombatActionBar = nil
  end
  
  if not self.tSettings.bEnable then
    return
  end
  
  local wndNonCombatActionBar = AlpacaWindowLibrary:CreateEmptyFrame(self.wndDisplay)  

  self:CreateNonCombatInnate(wndNonCombatActionBar)
  self:CreateNonCombatPotion(wndNonCombatActionBar)
  self:CreateNonCombatMount(wndNonCombatActionBar)
  self:CreateNonCombatPath(wndNonCombatActionBar)
  self:CreateNonCombatRecall(wndNonCombatActionBar)
  
  local nHorizontal = self.tSettings.tNonCombatActionBar.tPosition.nHorizontal
  local nVertical = self.tSettings.tNonCombatActionBar.tPosition.nVertical
  local nSpacing = self.tSettings.tNonCombatActionBar.nSpacing  
  
  self:ResizeContainer(wndNonCombatActionBar, nHorizontal, nVertical, nSpacing)
  
  self.wndNonCombatActionBar = wndNonCombatActionBar
end

function AlpacaActions:RefreshNonCombatActionBar()
  if not self.wndNonCombatActionBar then
    return
  end
  
  local bShow = true
  
  if self.tSettings.tNonCombatActionBar.bOnlyShowInCombat and not AlpacaGameLibrary:IsInCombatOrDamaged(self.unitPlayer) then
    bShow = false
  end
  
  if self.tSettings.bPreview then
    bShow = true
  end
  
  local wndNonCombatActionBar = self.wndNonCombatActionBar

  wndNonCombatActionBar:Show(bShow, true)
end

function AlpacaActions:OnButtonSignal(wndHandler, wndControl)
  if wndHandler ~= wndControl then
    return
  end

  local wndAction = AlpacaWindowLibrary:FindParent(wndHandler, "NonCombatAction")

  if not wndAction then
    return
  end

  local tData = wndAction:GetData()
  
  if not tData or not tData.eNonCombatActionType then
    return
  end
  
  if tData.eNonCombatActionType == keNonCombatActionType.Innate then
    AlpacaGameLibrary:SetInnate(wndHandler:GetData())
  elseif tData.eNonCombatActionType == keNonCombatActionType.Potion then
    AlpacaGameLibrary:SetPotion(wndHandler:GetData())
  elseif tData.eNonCombatActionType == keNonCombatActionType.Mount then
    AlpacaGameLibrary:SetMount(wndHandler:GetData())
  end
end

function AlpacaActions:OnGenerateTooltip(wndControl, wndHandler, eType, arg1, arg2)
  if Tooltip == nil then
    return
  end
 
  if eType == Tooltip.TooltipGenerateType_ItemInstance then
    if Tooltip.GetItemTooltipForm then
      Tooltip.GetItemTooltipForm(self, wndControl, arg1, {})
    end
  elseif eType == Tooltip.TooltipGenerateType_ItemData then
    if Tooltip.GetItemTooltipForm then
      Tooltip.GetItemTooltipForm(self, wndControl, arg1, {})
    end
  elseif eType == Tooltip.TooltipGenerateType_GameCommand then
		local xml = XmlDoc.new()
		xml:AddLine(arg2)
		wndControl:SetTooltipDoc(xml)
	elseif eType == Tooltip.TooltipGenerateType_Spell then
		if Tooltip.GetSpellTooltipForm ~= nil then
			Tooltip.GetSpellTooltipForm(self, wndControl, arg1)
		end
	end
end

function AlpacaActions:CreateNonCombatInnate(wndContainer)
  if self.wndNonCombatInnate then
    self.wndNonCombatInnate:Destroy()
    self.wndNonCombatInnate = nil
  end
  
  if not self.tSettings.tNonCombatActionBar.bShowInnate then
    return
  end
  
  local wndAction = Apollo.LoadForm(self.xmlDoc, "NonCombatAction", wndContainer, self)
  local wndButton = Apollo.LoadForm(self.xmlDoc, "GCButton", wndAction, self)
  local wndOptions = wndAction:FindChild("Options")
  
  wndButton:SetName("Button")
  wndButton:SetContentId(2) 
  wndAction:SetData({ eNonCombatActionType = keNonCombatActionType.Innate })
  
  self:CreateSpellOptions(wndOptions, AlpacaGameLibrary:GetInnateSpells())  
  
  self:ConfigureNonCombatAction(wndAction)
  
  self.wndNonCombatInnate = wndAction
end

function AlpacaActions:CreateNonCombatPotion(wndContainer)
  if self.wndNonCombatPotion then
    self.wndNonCombatPotion:Destroy()
    self.wndNonCombatPotion = nil
  end
  
  if not self.tSettings.tNonCombatActionBar.bShowPotion then
    return
  end
  
  local wndAction = Apollo.LoadForm(self.xmlDoc, "NonCombatAction", wndContainer, self)
  local wndButton = Apollo.LoadForm(self.xmlDoc, "GCButton", wndAction, self)
  local wndOptions = wndAction:FindChild("Options")
  
  wndButton:SetName("Button")
  wndButton:SetContentId(27) 
  wndAction:SetData({ eNonCombatActionType = keNonCombatActionType.Potion })
    
  self:CreateItemOptions(wndOptions, AlpacaGameLibrary:GetPotionItems())  
  
  self:ConfigureNonCombatAction(wndAction)
  
  self.wndNonCombatPotion = wndAction
end

function AlpacaActions:CreateNonCombatMount(wndContainer)
  if self.wndNonCombatMount then
    self.wndNonCombatMount:Destroy()
    self.wndNonCombatMount = nil
  end
  
  if not self.tSettings.tNonCombatActionBar.bShowMount then
    return
  end
  
  local wndAction = Apollo.LoadForm(self.xmlDoc, "NonCombatAction", wndContainer, self)
  local wndButton = Apollo.LoadForm(self.xmlDoc, "GCButton", wndAction, self)
  local wndOptions = wndAction:FindChild("Options")
  
  wndButton:SetName("Button")
  wndButton:SetContentId(26)
  wndAction:SetData({ eNonCombatActionType = keNonCombatActionType.Mount })
  
  self:CreateSpellOptions(wndOptions, AlpacaGameLibrary:GetMountSpells())  
  
  self:ConfigureNonCombatAction(wndAction)
  
  self.wndNonCombatMount = wndAction
end

function AlpacaActions:CreateNonCombatPath(wndContainer)
  if self.wndNonCombatPath then
    self.wndNonCombatPath:Destroy()
    self.wndNonCombatPath = nil
  end
  
  if not self.tSettings.tNonCombatActionBar.bShowPath then
    return
  end
  
  local wndAction = Apollo.LoadForm(self.xmlDoc, "NonCombatAction", wndContainer, self)
  local wndButton = Apollo.LoadForm(self.xmlDoc, "LASButton", wndAction, self)
  
  wndButton:SetName("Button")
  wndButton:SetContentId(9) 
  wndAction:SetData({ eNonCombatActionType = keNonCombatActionType.Path })
 
  self:ConfigureNonCombatAction(wndAction)
  
  self.wndNonCombatPath = wndAction
end

function AlpacaActions:CreateNonCombatRecall(wndContainer)
  if self.wndNonCombatRecall then
    self.wndNonCombatRecall:Destroy()
    self.wndNonCombatRecall = nil
  end
  
  if not self.tSettings.tNonCombatActionBar.bShowRecall then
    return
  end
  
  local wndAction = Apollo.LoadForm(self.xmlDoc, "NonCombatAction", wndContainer, self)
  local wndButton = Apollo.LoadForm(self.xmlDoc, "GCButton", wndAction, self)
  local wndOptions = wndAction:FindChild("Options")
  local nContentId = GameLib.GetDefaultRecallCommand()
    
  if nContentId == GameLib.CodeEnumRecallCommand.BindPoint and not GameLib.HasBindPoint() then
    nContentId = nil
  end
  
  wndButton:SetName("Button")
  
  if nContentId then
    wndButton:SetContentId(nContentId)
  else
    wndButton:Show(false, true)
  end

  wndAction:SetData({ eNonCombatActionType = keNonCombatActionType.Recall })

  self:CreateRecallOptions(wndOptions)
  
  self:ConfigureNonCombatAction(wndAction)
  
  self.wndNonCombatRecall = wndAction
end

function AlpacaActions:CreateRecallOptions(wndOptions)
  local arContentIds = {}
  
  if GameLib.HasBindPoint() then
		table.insert(arContentIds, GameLib.CodeEnumRecallCommand.BindPoint)
	end
  
  if HousingLib.IsResidenceOwner() then
		table.insert(arContentIds, GameLib.CodeEnumRecallCommand.House)
  end
 
  for _, objGuild in ipairs(GuildLib.GetGuilds()) do
    if objGuild:GetType() == GuildLib.GuildType_WarParty then
      table.insert(arContentIds, GameLib.CodeEnumRecallCommand.Warplot)
      break
    end
  end
  
	for idx, tSpell in pairs(AbilityBook.GetAbilitiesList(Spell.CodeEnumSpellTag.Misc) or {}) do
		if tSpell.bIsActive and tSpell.nId == GameLib.GetTeleportIlliumSpell():GetBaseSpellId() then
			table.insert(arContentIds, GameLib.CodeEnumRecallCommand.Illium)
		end
		if tSpell.bIsActive and tSpell.nId == GameLib.GetTeleportThaydSpell():GetBaseSpellId() then
      table.insert(arContentIds, GameLib.CodeEnumRecallCommand.Thayd)		
    end
	end

  for _, nContentId in ipairs(arContentIds) do
    local wndButton = Apollo.LoadForm(self.xmlDoc, "GCButton", wndOptions, self)
    wndButton:SetName("Option")
    wndButton:SetContentId(nContentId)
  end
end

-- Helpers

function AlpacaActions:CreateSpellOptions(wndOptions, arSpells)
  for _, objSpell in ipairs(arSpells) do
    local nSpellId = objSpell:GetId()
    local strIcon = objSpell:GetIcon()
    local wndButton = Apollo.LoadForm(self.xmlDoc, "Button", wndOptions, self)
    
    wndButton:SetName("Option")
    wndButton:FindChild("Icon"):SetSprite(strIcon)
    wndButton:SetData(nSpellId)
    
    if Tooltip and Tooltip.GetSpellTooltipForm then
      wndButton:SetTooltipDoc(nil)
      Tooltip.GetSpellTooltipForm(self, wndButton, objSpell, {})
		end
  end
end

function AlpacaActions:CreateItemOptions(wndOptions, arItems)
  for _, objItem in ipairs(arItems) do
    local nItemId = objItem:GetItemId()
    local strIcon = objItem:GetIcon()
    local wndButton = Apollo.LoadForm(self.xmlDoc, "Button", wndOptions, self)
    
    wndButton:SetName("Option")    
    wndButton:FindChild("Icon"):SetSprite(strIcon)
    wndButton:SetData(nItemId)
    
    if Tooltip and Tooltip.GetItemTooltipForm then
      wndButton:SetTooltipDoc(nil)
      Tooltip.GetItemTooltipForm(self, wndButton, objItem, {})
    end
  end
end

function AlpacaActions:ConfigureNonCombatAction(wndAction)
  local nIconSize = self.tSettings.tNonCombatActionBar.nIconSize

  wndAction:SetAnchorOffsets(0, 0, nIconSize, nIconSize)
  wndAction:FindChild("Button"):SetAnchorOffsets(0, 0, nIconSize, nIconSize)
    
  self:ConfigureNonCombatActionOptions(wndAction)
end
  
function AlpacaActions:ConfigureNonCombatActionOptions(wndAction)
  local wndOptions = wndAction:FindChild("Options")
  local wndOptionsButton = wndAction:FindChild("OptionsButton")
  local arChildren = wndOptions:GetChildren()
  
  if #arChildren < 2 then
    wndOptionsButton:Show(false, true)  
    return
  end

  local nIconSize = self.tSettings.tNonCombatActionBar.nIconSize
  
  for _, wndOption in ipairs(arChildren) do
    
    wndOption:SetAnchorOffsets(0, 0, nIconSize, nIconSize)
  end
    
  local nWidth = nIconSize + 16
  local nHeight = nIconSize * math.min(#arChildren, knMaxVisibleOptions)
  
  local nLeft = 0
  local nTop = -1 * (20 + nHeight)
  local nRight = nLeft + nWidth
  local nBottom = nTop + nHeight
  
  wndOptions:SetAnchorOffsets(nLeft, nTop, nRight, nBottom)
  wndOptions:ArrangeChildrenVert(0)
  
  wndOptionsButton:AttachWindow(wndOptions)
  wndOptionsButton:Show(true, true)    
end

function AlpacaActions:ResizeContainer(wndContainer, nHorizontal, nVertical, nSpacing) 
  if not wndContainer then
    return
  end
    
  local nContainerWidth = 0
  local nContainerHeight = 0

  for i, wndChild in ipairs(wndContainer:GetChildren()) do   
    local nWidth = wndChild:GetWidth()
    local nHeight = wndChild:GetHeight()
   
    local nLeft = (i - 1) * (nWidth + nSpacing)
    local nTop = 0
    local nRight = nLeft + nWidth
    local nBottom = nHeight
  
    wndChild:SetAnchorOffsets(nLeft, nTop, nRight, nBottom)
    
    nContainerWidth = math.max(nContainerWidth, nRight)
    nContainerHeight = math.max(nContainerHeight, nBottom)
  end

  local nLeft = nHorizontal - math.ceil(nContainerWidth / 2)
  local nTop = nVertical - math.ceil(nContainerHeight / 2)
  local nRight = nLeft + nContainerWidth
  local nBottom = nTop + nContainerHeight
  
  wndContainer:SetAnchorOffsets(nLeft, nTop, nRight, nBottom)
end

-- Start

local AlpacaActionsInst = AlpacaActions:new()
AlpacaActions:Init()
