require "Apollo"
require "GameLib"
require "Unit"
require "XmlDoc"

local AlpacaGameLibrary = nil
local AlpacaPluginLibrary = nil
local AlpacaStandardLibrary = nil
local AlpacaWindowLibrary = nil

local AlpacaPlates = {} 

-- Constants

local kfRefreshFullDelay = 1
local kfRefreshVisibilityDelay = 0.4
local kfLogStatsDelay = 0.25

local knNameplatePoolSize = 50
local knMaxRange = 150

local keColorSet = {
  UnitType = 1,
  UnitFlags = 2,
  UnitDifficulty = 3
}

local ktLookupClassIcon = {
	[GameLib.CodeEnumClass.Engineer] 	  	= "AlpacaSprites:EngineerIcon",
	[GameLib.CodeEnumClass.Esper] 		  	= "AlpacaSprites:EsperIcon",
	[GameLib.CodeEnumClass.Medic] 			  = "AlpacaSprites:MedicIcon",
	[GameLib.CodeEnumClass.Spellslinger]	= "AlpacaSprites:SpellslingerIcon",
	[GameLib.CodeEnumClass.Stalker] 	  	= "AlpacaSprites:StalkerIcon",
	[GameLib.CodeEnumClass.Warrior] 	  	= "AlpacaSprites:WarriorIcon"
}

local ktLookupRankIcon = {
 	[Unit.CodeEnumRank.Elite]             = "AlpacaSprites:NonPlayerEliteIcon",
  [Unit.CodeEnumRank.Superior]          = "AlpacaSprites:NonPlayerSuperiorIcon",
  [Unit.CodeEnumRank.Champion]          = "AlpacaSprites:NonPlayerChampionIcon",
  [Unit.CodeEnumRank.Standard]          = "AlpacaSprites:NonPlayerStandardIcon"
}

-- Plugin

function AlpacaPlates:PluginId()
  return "AlpacaPlates"
end

function AlpacaPlates:PluginName()
  return "Nameplates"
end

function AlpacaPlates:PluginDefaults() 
  local fnCreateColor = function(...)
    return AlpacaWindowLibrary:AlpacaColor(...)
  end
 
  local fnCreateDefaultStyle = function() 
    return {
      nRange = knMaxRange,
      tName = {
        bEnable = true,
        eColorSet = keColorSet.UnitType,
        tPosition = { nHorizontal = 150, nVertical = 130 },
        strAlign = "Center"
      },
      tLevel = {
        bEnable = true,
        eColorSet = keColorSet.UnitDifficulty,
        tPosition = { nHorizontal = 90, nVertical = 150 },
        strAlign = "Center"
      },
      tIcon = {
        bEnable = true,
        bShowPlayerIcons = true,
        bShowNonPlayerEliteIcons = true,
        bShowNonPlayerCommonIcons = false,
        eColorSet = keColorSet.UnitType,
        tPosition = { nHorizontal = 60, nVertical = 120 },
        nSize = 40
      },
      tInterruptArmor = {
        bEnable = true,
        eColorSet = keColorSet.UnitType,
        tPosition = { nHorizontal = 200, nVertical = 120 },
        nSize = 40
      },
      tRewards = {
        bEnable = true,
        tPosition = { nHorizontal = 150, nVertical = 100 },
        strAlign = "Center"
      },
      tHealthBar = {
        bEnable = true,
        tPosition = { nHorizontal = 100, nVertical = 145 },
        tSize = { nWidth = 100, nHeight = 10 },
        eColorSet = keColorSet.UnitFlags
      },
      tShieldBar = {
        bEnable = true,
        tPosition = { nHorizontal = 100, nVertical = 140 },
        tSize = { nWidth = 100, nHeight = 5 },
        crColor = fnCreateColor(0, 0, 1, 0.7)
      },
      tAbsorbBar = {
        bEnable = true,
        tPosition = { nHorizontal = 100, nVertical = 115 },
        tSize = { nWidth = 100, nHeight = 5 },
        crColor = fnCreateColor(1, 0.5, 0, 0.7)
      },
      tCastBar = {
        bEnable = false,
        tPosition = { nHorizontal = 100, nVertical = 110 },
        tSize = { nWidth = 100, nHeight = 5 },
        crColor = fnCreateColor(1, 1, 0, 0.7)
      },
      tDisableBar = {
        bEnable = true,
        tPosition = { nHorizontal = 100, nVertical = 110 },
        tSize = { nWidth = 100, nHeight = 5 },
        crColor = fnCreateColor(0.8, 0, 0.8, 0.7),
        bEnableBlind = true,
        bEnableDisarm = true,
        bEnableKnockdown = true,
        bEnableStun = true,
        bEnableSubdue = true,
        bEnableTether = true
      }
    }    
  end
  
  local fnCreateDefaultPetStyle = function() 
    local tStyle = fnCreateDefaultStyle()
    
    tStyle.tName.bEnable = false
    tStyle.tCastBar.bEnable = false

    return tStyle
  end
  
  local fnCreateDefaultSimpleStyle = function() 
    local tStyle = fnCreateDefaultStyle()
    
    tStyle.tName.tPosition.nVertical = 150
    tStyle.tLevel.bEnable = false
    tStyle.tIcon.bEnable = false
    tStyle.tInterruptArmor.bEnable = false
    tStyle.tHealthBar.bEnable = false
    tStyle.tShieldBar.bEnable = false
    tStyle.tAbsorbBar.bEnable = false
    tStyle.tCastBar.bEnable = false
    
    return tStyle
  end
  
  return {
    bEnable = true,
    bHideWhileChatBubbleActive = true,
    tUserPlayer = {
      nStyle = 1,
      nStyleInCombat = 1
    },
    tTarget = {
      nStyle = 1,
      nStyleInCombat = 1
    },
    tPlayerFriendly = {
      nStyle = 1,
      nStyleInCombat = 1
    },
    tPlayerHostile = {
      nStyle = 1,
      nStyleInCombat = 1
    },
    tPetFriendly = {
      nStyle = 2,
      nStyleInCombat = 2
    },
    tPetHostile = {
      nStyle = 2,
      nStyleInCombat = 2
    },
    tNonPlayerFriendly = {
      nStyle = 3,
      nStyleInCombat = 1
    },
    tNonPlayerHostile = {
      nStyle = 1,
      nStyleInCombat = 1
    },
    tNonPlayerNeutral = {
      nStyle = 1,
      nStyleInCombat = 1
    },
    tColorSets = {
      tUnitType = {
        [AlpacaGameLibrary.EnumUnitType.Engineer]              = fnCreateColor(1, 0.8, 0, 0.9),
        [AlpacaGameLibrary.EnumUnitType.Esper]                 = fnCreateColor(1, 0, 1, 0.9),
        [AlpacaGameLibrary.EnumUnitType.Medic]                 = fnCreateColor(1, 1, 1, 0.9),
        [AlpacaGameLibrary.EnumUnitType.Spellslinger]          = fnCreateColor(0, 1, 1, 0.9),
        [AlpacaGameLibrary.EnumUnitType.Stalker]               = fnCreateColor(1, 1, 0, 0.9),
        [AlpacaGameLibrary.EnumUnitType.Warrior]               = fnCreateColor(1, 0.77, 0.65, 0.9),
        [AlpacaGameLibrary.EnumUnitType.Pet]                   = fnCreateColor(0.8, 0.8, 0.8, 0.7),
        [AlpacaGameLibrary.EnumUnitType.NonPlayerFriendly]     = fnCreateColor(0, 1, 0, 0.7),
        [AlpacaGameLibrary.EnumUnitType.NonPlayerNeutral]      = fnCreateColor(1, 1, 0, 0.7),
        [AlpacaGameLibrary.EnumUnitType.NonPlayerHostile]      = fnCreateColor(1, 0, 0, 0.7)
      },
      tUnitFlags = {
        [AlpacaGameLibrary.EnumUnitFlags.PlayerFriendly]              = fnCreateColor(0, 1, 0, 0.7),
        [AlpacaGameLibrary.EnumUnitFlags.PlayerFriendlyPvp]           = fnCreateColor(0, 1, 0, 0.7),
        [AlpacaGameLibrary.EnumUnitFlags.PlayerHostile]               = fnCreateColor(1, 1, 0, 0.7),      
        [AlpacaGameLibrary.EnumUnitFlags.PlayerHostilePvp]            = fnCreateColor(1, 0, 0, 0.7),
        [AlpacaGameLibrary.EnumUnitFlags.PlayerGroupMember]           = fnCreateColor(0, 1, 0, 0.7),
        [AlpacaGameLibrary.EnumUnitFlags.PlayerGroupMemberPvp]        = fnCreateColor(0, 1, 0, 0.7),
        [AlpacaGameLibrary.EnumUnitFlags.PetFriendly]                 = fnCreateColor(0, 1, 0, 0.7),
        [AlpacaGameLibrary.EnumUnitFlags.PetFriendlyPvp]              = fnCreateColor(0, 1, 0, 0.7),
        [AlpacaGameLibrary.EnumUnitFlags.PetHostile]                  = fnCreateColor(1, 1, 0, 0.7),
        [AlpacaGameLibrary.EnumUnitFlags.PetHostilePvp]               = fnCreateColor(1, 0, 0, 0.7),
        [AlpacaGameLibrary.EnumUnitFlags.NonPlayerFriendly]           = fnCreateColor(0, 1, 0, 0.7),
        [AlpacaGameLibrary.EnumUnitFlags.NonPlayerNeutral]            = fnCreateColor(1, 1, 0, 0.7),
        [AlpacaGameLibrary.EnumUnitFlags.NonPlayerHostile]            = fnCreateColor(1, 0, 0, 0.7)
      },
      tUnitDifficulty = {
        [AlpacaGameLibrary.EnumUnitDifficulty.Trivial]      = fnCreateColor(0.7, 0.7, 0.7, 0.7),
        [AlpacaGameLibrary.EnumUnitDifficulty.Inferior]     = fnCreateColor(0, 1, 0, 0.7),
        [AlpacaGameLibrary.EnumUnitDifficulty.Minor]        = fnCreateColor(0.3, 0.5, 1, 0.7),
        [AlpacaGameLibrary.EnumUnitDifficulty.Easy]         = fnCreateColor(0, 1, 1, 0.7),
        [AlpacaGameLibrary.EnumUnitDifficulty.Average]      = fnCreateColor(1, 1, 1, 0.7),
        [AlpacaGameLibrary.EnumUnitDifficulty.Moderate]    	= fnCreateColor(1, 1, 0, 0.7),
        [AlpacaGameLibrary.EnumUnitDifficulty.Tough]        = fnCreateColor(1, 0.5, 0, 0.7),
        [AlpacaGameLibrary.EnumUnitDifficulty.Hard]         = fnCreateColor(1, 0, 0, 0.7),
        [AlpacaGameLibrary.EnumUnitDifficulty.Impossible]   = fnCreateColor(1, 0, 1, 0.7)
      }
    },    
    arStyles = {
      fnCreateDefaultStyle(),
      fnCreateDefaultPetStyle(),
      fnCreateDefaultSimpleStyle(),
      fnCreateDefaultStyle(),
      fnCreateDefaultStyle()
    }
  }
end

function AlpacaPlates:PluginForm()
  local fnCreateStyleOptions = function ()
    return {
      { strLabel = "Disabled", nValue = 0 },
      { strLabel = "Style 1", nValue = 1 },
      { strLabel = "Style 2", nValue = 2 },
      { strLabel = "Style 3", nValue = 3 },
      { strLabel = "Style 4", nValue = 4 },
      { strLabel = "Style 5", nValue = 5 }
    }   
  end
  
  local fnCreateColorSetOptions = function()
    return {
      { strLabel = "Unit Type", nValue = keColorSet.UnitType },
      { strLabel = "Unit Flags", nValue = keColorSet.UnitFlags },
      { strLabel = "Unit Difficulty", nValue = keColorSet.UnitDifficulty }
    }
  end
  
  local fnCreateAlignOptions = function ()
    return {
      { strLabel = "Left", nValue = "Left" },
      { strLabel = "Center", nValue = "Center" },
      { strLabel = "Right", nValue = "Right" }
    }
  end
  
  local fnCreateStyleSubForm = function(nStyle)
    return AlpacaPluginLibrary:CreateSubForm("Style " .. nStyle, {
      AlpacaPluginLibrary:CreateHeader("Style " .. nStyle),
      AlpacaPluginLibrary:CreateSlider("Max Range",               { "arStyles", nStyle, "nRange" }, 0, knMaxRange, 1),
      
      AlpacaPluginLibrary:CreateSubHeader("Name"),
      AlpacaPluginLibrary:CreateCheckbox("Enable",                { "arStyles", nStyle, "tName", "bEnable" }),
      AlpacaPluginLibrary:CreateSelect("Color Set",               { "arStyles", nStyle, "tName", "eColorSet" }, fnCreateColorSetOptions()),
      AlpacaPluginLibrary:CreateSelect("Align",                   { "arStyles", nStyle, "tName", "strAlign" }, fnCreateAlignOptions()),
      AlpacaPluginLibrary:CreateSlider("Position (Horizontal)",   { "arStyles", nStyle, "tName", "tPosition", "nHorizontal" }, 0, 300, 1),
      AlpacaPluginLibrary:CreateSlider("Position (Vertical)",     { "arStyles", nStyle, "tName", "tPosition", "nVertical" }, 0, 300, 1),
      
      AlpacaPluginLibrary:CreateSubHeader("Level"),
      AlpacaPluginLibrary:CreateCheckbox("Enable",                { "arStyles", nStyle, "tLevel", "bEnable" }),
      AlpacaPluginLibrary:CreateSelect("Color Set",               { "arStyles", nStyle, "tLevel", "eColorSet" }, fnCreateColorSetOptions()),
      AlpacaPluginLibrary:CreateSelect("Align",                   { "arStyles", nStyle, "tLevel", "strAlign" }, fnCreateAlignOptions()),
      AlpacaPluginLibrary:CreateSlider("Position (Horizontal)",   { "arStyles", nStyle, "tLevel", "tPosition", "nHorizontal" }, 0, 300, 1),
      AlpacaPluginLibrary:CreateSlider("Position (Vertical)",     { "arStyles", nStyle, "tLevel", "tPosition", "nVertical" }, 0, 300, 1),
      
      AlpacaPluginLibrary:CreateSubHeader("Icon"),
      AlpacaPluginLibrary:CreateCheckbox("Enable",                { "arStyles", nStyle, "tIcon", "bEnable" }),
      AlpacaPluginLibrary:CreateCheckbox("Player Icons",          { "arStyles", nStyle, "tIcon", "bShowPlayerIcons" }),
      AlpacaPluginLibrary:CreateCheckbox("NPC (Elite) Icons",     { "arStyles", nStyle, "tIcon", "bShowNonPlayerEliteIcons" }),
      AlpacaPluginLibrary:CreateCheckbox("NPC (Common) Icons",    { "arStyles", nStyle, "tIcon", "bShowNonPlayerCommonIcons" }),
      AlpacaPluginLibrary:CreateSelect("Color Set",               { "arStyles", nStyle, "tIcon", "eColorSet" }, fnCreateColorSetOptions()),
      AlpacaPluginLibrary:CreateSlider("Position (Horizontal)",   { "arStyles", nStyle, "tIcon", "tPosition", "nHorizontal" }, 0, 300, 1),
      AlpacaPluginLibrary:CreateSlider("Position (Vertical)",     { "arStyles", nStyle, "tIcon", "tPosition", "nVertical" }, 0, 300, 1),
      AlpacaPluginLibrary:CreateSlider("Size",                    { "arStyles", nStyle, "tIcon", "nSize" }, 0, 300, 1),
      
      AlpacaPluginLibrary:CreateSubHeader("Interrupt Armor"),
      AlpacaPluginLibrary:CreateCheckbox("Enable",                { "arStyles", nStyle, "tInterruptArmor", "bEnable" }),
      AlpacaPluginLibrary:CreateSelect("Color Set",               { "arStyles", nStyle, "tInterruptArmor", "eColorSet" }, fnCreateColorSetOptions()),
      AlpacaPluginLibrary:CreateSlider("Position (Horizontal)",   { "arStyles", nStyle, "tInterruptArmor", "tPosition", "nHorizontal" }, 0, 300, 1),
      AlpacaPluginLibrary:CreateSlider("Position (Vertical)",     { "arStyles", nStyle, "tInterruptArmor", "tPosition", "nVertical" }, 0, 300, 1),
      AlpacaPluginLibrary:CreateSlider("Size",                    { "arStyles", nStyle, "tInterruptArmor", "nSize" }, 0, 300, 1),
      
      AlpacaPluginLibrary:CreateSubHeader("Rewards"),
      AlpacaPluginLibrary:CreateCheckbox("Enable",                { "arStyles", nStyle, "tRewards", "bEnable" }),
      AlpacaPluginLibrary:CreateSelect("Align",                   { "arStyles", nStyle, "tRewards", "strAlign" }, fnCreateAlignOptions()),
      AlpacaPluginLibrary:CreateSlider("Position (Horizontal)",   { "arStyles", nStyle, "tRewards", "tPosition", "nHorizontal" }, 0, 300, 1),
      AlpacaPluginLibrary:CreateSlider("Position (Vertical)",     { "arStyles", nStyle, "tRewards", "tPosition", "nVertical" }, 0, 300, 1),

      AlpacaPluginLibrary:CreateSubHeader("Health Bar"),
      AlpacaPluginLibrary:CreateCheckbox("Enable",                { "arStyles", nStyle, "tHealthBar", "bEnable" }),
      AlpacaPluginLibrary:CreateSelect("Color Set",               { "arStyles", nStyle, "tHealthBar", "eColorSet" }, fnCreateColorSetOptions()),
      AlpacaPluginLibrary:CreateSlider("Position (Horizontal)",   { "arStyles", nStyle, "tHealthBar", "tPosition", "nHorizontal" }, 0, 300, 1),
      AlpacaPluginLibrary:CreateSlider("Position (Vertical)",     { "arStyles", nStyle, "tHealthBar", "tPosition", "nVertical" }, 0, 300, 1),
      AlpacaPluginLibrary:CreateSlider("Size (Width)",            { "arStyles", nStyle, "tHealthBar", "tSize", "nWidth" }, 0, 300, 1),
      AlpacaPluginLibrary:CreateSlider("Size (Height)",           { "arStyles", nStyle, "tHealthBar", "tSize", "nHeight" }, 0, 300, 1),

      AlpacaPluginLibrary:CreateSubHeader("Shield Bar"),
      AlpacaPluginLibrary:CreateCheckbox("Enable",                { "arStyles", nStyle, "tShieldBar", "bEnable" }),
      AlpacaPluginLibrary:CreateColor("Color",                    { "arStyles", nStyle, "tShieldBar", "crColor" }),
      AlpacaPluginLibrary:CreateSlider("Position (Horizontal)",   { "arStyles", nStyle, "tShieldBar", "tPosition", "nHorizontal" }, 0, 300, 1),
      AlpacaPluginLibrary:CreateSlider("Position (Vertical)",     { "arStyles", nStyle, "tShieldBar", "tPosition", "nVertical" }, 0, 300, 1),
      AlpacaPluginLibrary:CreateSlider("Size (Width)",            { "arStyles", nStyle, "tShieldBar", "tSize", "nWidth" }, 0, 300, 1),
      AlpacaPluginLibrary:CreateSlider("Size (Height)",           { "arStyles", nStyle, "tShieldBar", "tSize", "nHeight" }, 0, 300, 1),

      AlpacaPluginLibrary:CreateSubHeader("Absorb Bar"),
      AlpacaPluginLibrary:CreateCheckbox("Enable",                { "arStyles", nStyle, "tAbsorbBar", "bEnable" }),
      AlpacaPluginLibrary:CreateColor("Color",                    { "arStyles", nStyle, "tAbsorbBar", "crColor" }),
      AlpacaPluginLibrary:CreateSlider("Position (Horizontal)",   { "arStyles", nStyle, "tAbsorbBar", "tPosition", "nHorizontal" }, 0, 300, 1),
      AlpacaPluginLibrary:CreateSlider("Position (Vertical)",     { "arStyles", nStyle, "tAbsorbBar", "tPosition", "nVertical" }, 0, 300, 1),
      AlpacaPluginLibrary:CreateSlider("Size (Width)",            { "arStyles", nStyle, "tAbsorbBar", "tSize", "nWidth" }, 0, 300, 1),
      AlpacaPluginLibrary:CreateSlider("Size (Height)",           { "arStyles", nStyle, "tAbsorbBar", "tSize", "nHeight" }, 0, 300, 1),

      AlpacaPluginLibrary:CreateSubHeader("Cast Bar"),
      AlpacaPluginLibrary:CreateCheckbox("Enable",                { "arStyles", nStyle, "tCastBar", "bEnable" }),
      AlpacaPluginLibrary:CreateColor("Color",                    { "arStyles", nStyle, "tCastBar", "crColor" }),
      AlpacaPluginLibrary:CreateSlider("Position (Horizontal)",   { "arStyles", nStyle, "tCastBar", "tPosition", "nHorizontal" }, 0, 300, 1),
      AlpacaPluginLibrary:CreateSlider("Position (Vertical)",     { "arStyles", nStyle, "tCastBar", "tPosition", "nVertical" }, 0, 300, 1),
      AlpacaPluginLibrary:CreateSlider("Size (Width)",            { "arStyles", nStyle, "tCastBar", "tSize", "nWidth" }, 0, 300, 1),
      AlpacaPluginLibrary:CreateSlider("Size (Height)",           { "arStyles", nStyle, "tCastBar", "tSize", "nHeight" }, 0, 300, 1),
      
      AlpacaPluginLibrary:CreateSubHeader("Disable Bar"),
      AlpacaPluginLibrary:CreateCheckbox("Enable",                { "arStyles", nStyle, "tDisableBar", "bEnable" }),
      AlpacaPluginLibrary:CreateColor("Color",                    { "arStyles", nStyle, "tDisableBar", "crColor" }),
      AlpacaPluginLibrary:CreateSlider("Position (Horizontal)",   { "arStyles", nStyle, "tDisableBar", "tPosition", "nHorizontal" }, 0, 300, 1),
      AlpacaPluginLibrary:CreateSlider("Position (Vertical)",     { "arStyles", nStyle, "tDisableBar", "tPosition", "nVertical" }, 0, 300, 1),
      AlpacaPluginLibrary:CreateSlider("Size (Width)",            { "arStyles", nStyle, "tDisableBar", "tSize", "nWidth" }, 0, 300, 1),
      AlpacaPluginLibrary:CreateSlider("Size (Height)",           { "arStyles", nStyle, "tDisableBar", "tSize", "nHeight" }, 0, 300, 1),
    })
  end
  
  local arInputDefinitions = {
    AlpacaPluginLibrary:CreateHeader("Nameplates"),
    AlpacaPluginLibrary:CreateCheckbox("Enable", { "bEnable" }),
    AlpacaPluginLibrary:CreateCheckbox("Hide While Chat Bubble Active", { "bHideWhileChatBubbleActive" }),
    
    AlpacaPluginLibrary:CreateSubHeader("Current Player"),
    AlpacaPluginLibrary:CreateSelect("Style",                 { "tUserPlayer", "nStyle" }, fnCreateStyleOptions()),
    AlpacaPluginLibrary:CreateSelect("Style (Combat)",        { "tUserPlayer", "nStyleInCombat" }, fnCreateStyleOptions()),
      
    AlpacaPluginLibrary:CreateSubHeader("Target"),
    AlpacaPluginLibrary:CreateSelect("Style",                 { "tTarget", "nStyle" }, fnCreateStyleOptions()),
    AlpacaPluginLibrary:CreateSelect("Style (Combat)",        { "tTarget", "nStyleInCombat" }, fnCreateStyleOptions()),
      
    AlpacaPluginLibrary:CreateSubHeader("Friendly Player"),
    AlpacaPluginLibrary:CreateSelect("Style",                 { "tPlayerFriendly", "nStyle" }, fnCreateStyleOptions()),
    AlpacaPluginLibrary:CreateSelect("Style (Combat)",        { "tPlayerFriendly", "nStyleInCombat" }, fnCreateStyleOptions()),
      
    AlpacaPluginLibrary:CreateSubHeader("Hostile Player"),
    AlpacaPluginLibrary:CreateSelect("Style",                 { "tPlayerHostile", "nStyle" }, fnCreateStyleOptions()),
    AlpacaPluginLibrary:CreateSelect("Style (Combat)",        { "tPlayerHostile", "nStyleInCombat" }, fnCreateStyleOptions()),
      
    AlpacaPluginLibrary:CreateSubHeader("Friendly Pet"),
    AlpacaPluginLibrary:CreateSelect("Style",                 { "tPetFriendly", "nStyle" }, fnCreateStyleOptions()),
    AlpacaPluginLibrary:CreateSelect("Style (Combat)",        { "tPetFriendly", "nStyleInCombat" }, fnCreateStyleOptions()),
      
    AlpacaPluginLibrary:CreateSubHeader("Hostile Pet"),
    AlpacaPluginLibrary:CreateSelect("Style",                 { "tPetHostile", "nStyle" }, fnCreateStyleOptions()),
    AlpacaPluginLibrary:CreateSelect("Style (Combat)",        { "tPetHostile", "nStyleInCombat" }, fnCreateStyleOptions()),
      
    AlpacaPluginLibrary:CreateSubHeader("Friendly NPC"),
    AlpacaPluginLibrary:CreateSelect("Style",                 { "tNonPlayerFriendly", "nStyle" }, fnCreateStyleOptions()),
    AlpacaPluginLibrary:CreateSelect("Style (Combat)",        { "tNonPlayerFriendly", "nStyleInCombat" }, fnCreateStyleOptions()),
      
    AlpacaPluginLibrary:CreateSubHeader("Neutral NPC"),
    AlpacaPluginLibrary:CreateSelect("Style",                 { "tNonPlayerNeutral", "nStyle" }, fnCreateStyleOptions()),
    AlpacaPluginLibrary:CreateSelect("Style (Combat)",        { "tNonPlayerNeutral", "nStyleInCombat" }, fnCreateStyleOptions()),
      
    AlpacaPluginLibrary:CreateSubHeader("Hostile NPC"),
    AlpacaPluginLibrary:CreateSelect("Style",                 { "tNonPlayerHostile", "nStyle" }, fnCreateStyleOptions()),
    AlpacaPluginLibrary:CreateSelect("Style (Combat)",        { "tNonPlayerHostile", "nStyleInCombat" }, fnCreateStyleOptions())
  }
  
  local arSubForms = {
    AlpacaPluginLibrary:CreateSubForm("Color Sets", {
      AlpacaPluginLibrary:CreateHeader("Color Sets"),
      
      AlpacaPluginLibrary:CreateSubHeader("Unit Type"),
      AlpacaPluginLibrary:CreateColor("Engineer",     { "tColorSets", "tUnitType", AlpacaGameLibrary.EnumUnitType.Engineer }),
      AlpacaPluginLibrary:CreateColor("Esper",        { "tColorSets", "tUnitType", AlpacaGameLibrary.EnumUnitType.Esper }),
      AlpacaPluginLibrary:CreateColor("Medic",        { "tColorSets", "tUnitType", AlpacaGameLibrary.EnumUnitType.Medic }),
      AlpacaPluginLibrary:CreateColor("Spellslinger", { "tColorSets", "tUnitType", AlpacaGameLibrary.EnumUnitType.Spellslinger }),
      AlpacaPluginLibrary:CreateColor("Stalker",      { "tColorSets", "tUnitType", AlpacaGameLibrary.EnumUnitType.Stalker }),
      AlpacaPluginLibrary:CreateColor("Warrior",      { "tColorSets", "tUnitType", AlpacaGameLibrary.EnumUnitType.Warrior }),
      AlpacaPluginLibrary:CreateColor("Pet",          { "tColorSets", "tUnitType", AlpacaGameLibrary.EnumUnitType.Pet }),
      AlpacaPluginLibrary:CreateColor("Friendly NPC", { "tColorSets", "tUnitType", AlpacaGameLibrary.EnumUnitType.NonPlayerFriendly }),
      AlpacaPluginLibrary:CreateColor("Neutral NPC",  { "tColorSets", "tUnitType", AlpacaGameLibrary.EnumUnitType.NonPlayerNeutral }),
      AlpacaPluginLibrary:CreateColor("Hostile NPC",  { "tColorSets", "tUnitType", AlpacaGameLibrary.EnumUnitType.NonPlayerHostile }),
   
      AlpacaPluginLibrary:CreateSubHeader("Unit Flags"),
      AlpacaPluginLibrary:CreateColor("Group Member",           { "tColorSets", "tUnitFlags", AlpacaGameLibrary.EnumUnitFlags.PlayerGroupMember }),
      AlpacaPluginLibrary:CreateColor("Group Member (PvP)",     { "tColorSets", "tUnitFlags", AlpacaGameLibrary.EnumUnitFlags.PlayerGroupMemberPvp }),
      AlpacaPluginLibrary:CreateColor("Friendly Player",        { "tColorSets", "tUnitFlags", AlpacaGameLibrary.EnumUnitFlags.PlayerFriendly }),
      AlpacaPluginLibrary:CreateColor("Friendly Player (PvP)",  { "tColorSets", "tUnitFlags", AlpacaGameLibrary.EnumUnitFlags.PlayerFriendlyPvp }),
      AlpacaPluginLibrary:CreateColor("Hostile Player",         { "tColorSets", "tUnitFlags", AlpacaGameLibrary.EnumUnitFlags.PlayerHostile }),
      AlpacaPluginLibrary:CreateColor("Hostile Player (PvP)",   { "tColorSets", "tUnitFlags", AlpacaGameLibrary.EnumUnitFlags.PlayerHostilePvp }),
      AlpacaPluginLibrary:CreateColor("Friendly Pet",           { "tColorSets", "tUnitFlags", AlpacaGameLibrary.EnumUnitFlags.PetFriendly }),
      AlpacaPluginLibrary:CreateColor("Friendly Pet (PvP)",     { "tColorSets", "tUnitFlags", AlpacaGameLibrary.EnumUnitFlags.PetFriendlyPvp }),
      AlpacaPluginLibrary:CreateColor("Hostile Pet",            { "tColorSets", "tUnitFlags", AlpacaGameLibrary.EnumUnitFlags.PetHostile }),
      AlpacaPluginLibrary:CreateColor("Hostile Pet (PvP)",      { "tColorSets", "tUnitFlags", AlpacaGameLibrary.EnumUnitFlags.PetHostilePvp }),
      AlpacaPluginLibrary:CreateColor("Friendly NPC",           { "tColorSets", "tUnitFlags", AlpacaGameLibrary.EnumUnitFlags.NonPlayerFriendly }),
      AlpacaPluginLibrary:CreateColor("Neutral NPC",            { "tColorSets", "tUnitFlags", AlpacaGameLibrary.EnumUnitFlags.NonPlayerNeutral }),
      AlpacaPluginLibrary:CreateColor("Hostile NPC",            { "tColorSets", "tUnitFlags", AlpacaGameLibrary.EnumUnitFlags.NonPlayerHostile }),
        
      AlpacaPluginLibrary:CreateSubHeader("Unit Difficulty"),
      AlpacaPluginLibrary:CreateColor("Trivial",      { "tColorSets", "tUnitDifficulty", AlpacaGameLibrary.EnumUnitDifficulty.Trivial }),
      AlpacaPluginLibrary:CreateColor("Inferior",     { "tColorSets", "tUnitDifficulty", AlpacaGameLibrary.EnumUnitDifficulty.Inferior }),
      AlpacaPluginLibrary:CreateColor("Minor",        { "tColorSets", "tUnitDifficulty", AlpacaGameLibrary.EnumUnitDifficulty.Minor }),
      AlpacaPluginLibrary:CreateColor("Easy",         { "tColorSets", "tUnitDifficulty", AlpacaGameLibrary.EnumUnitDifficulty.Easy }),
      AlpacaPluginLibrary:CreateColor("Average",      { "tColorSets", "tUnitDifficulty", AlpacaGameLibrary.EnumUnitDifficulty.Average }),
      AlpacaPluginLibrary:CreateColor("Moderate",     { "tColorSets", "tUnitDifficulty", AlpacaGameLibrary.EnumUnitDifficulty.Moderate }),
      AlpacaPluginLibrary:CreateColor("Tough",        { "tColorSets", "tUnitDifficulty", AlpacaGameLibrary.EnumUnitDifficulty.Tough }),
      AlpacaPluginLibrary:CreateColor("Hard",         { "tColorSets", "tUnitDifficulty", AlpacaGameLibrary.EnumUnitDifficulty.Hard }),
      AlpacaPluginLibrary:CreateColor("Impossible",   { "tColorSets", "tUnitDifficulty", AlpacaGameLibrary.EnumUnitDifficulty.Impossible }),
    })
   }
  
  for i = 1, 5 do
    table.insert(arSubForms, fnCreateStyleSubForm(i))
  end
  
  return AlpacaPluginLibrary:CreateForm(arInputDefinitions, arSubForms)
end

function AlpacaPlates:OnPluginRegistered(wndDisplay)
  self.wndDisplay = wndDisplay
end

function AlpacaPlates:OnPluginSettingsChanged(tSettings)
  self.tSettings = tSettings
  self.fLastRefreshFull = nil
end

-- Addon

function AlpacaPlates:new(o)
  o = o or {}
  setmetatable(o, self)
  self.__index = self 
  return o
end

function AlpacaPlates:Init()
	self.arPendingUnitIds = {}
  self.tActiveNameplates = {}
  self.arAvailableNameplates = {}

  local bHasConfigureFunction = false
  local strConfigureButtonText = nil
  local tDependencies = { "AlpacaGameLibrary", "AlpacaPluginLibrary", "AlpacaStandardLibrary", "AlpacaWindowLibrary", "RewardIcons" }
  
  Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end
 
function AlpacaPlates:OnLoad()
  Apollo.LoadSprites("AlpacaSprites.xml")

  AlpacaGameLibrary = Apollo.GetPackage("AlpacaGameLibrary").tPackage
  AlpacaPluginLibrary = Apollo.GetPackage("AlpacaPluginLibrary").tPackage
  AlpacaStandardLibrary = Apollo.GetPackage("AlpacaStandardLibrary").tPackage
  AlpacaWindowLibrary = Apollo.GetPackage("AlpacaWindowLibrary").tPackage

  Apollo.RegisterEventHandler("VarChange_FrameCount", "OnFrame", self)

  -- events    
  Apollo.RegisterEventHandler("UnitCreated",              "OnUnitCreated", self)
  Apollo.RegisterEventHandler("UnitDestroyed",            "OnUnitDestroyed", self)
  Apollo.RegisterEventHandler("UnitTextBubbleCreate",	    "OnUnitTextBubbleToggled", self)
	Apollo.RegisterEventHandler("UnitTextBubblesDestroyed", "OnUnitTextBubbleToggled", self)
	Apollo.RegisterEventHandler("UnitNameChanged",          "OnUnitNameChanged", self)
	Apollo.RegisterEventHandler("UnitLevelChanged",         "OnUnitLevelChanged", self)
  Apollo.RegisterEventHandler("UnitEnteredCombat",        "OnUnitEnteredCombat", self)
  Apollo.RegisterEventHandler("TargetUnitChanged",        "OnTargetUnitChanged", self)

	local arRewardEvents = {
		"QuestObjectiveUpdated", 
    "QuestStateChanged", 
    "ChallengeAbandon", 
    "ChallengeLeftArea",
		"ChallengeFailTime", 
    "ChallengeFailArea", 
    "ChallengeActivate", 
    "ChallengeCompleted",
		"ChallengeFailGeneric", 
    "PublicEventObjectiveUpdate", 
    "PublicEventUnitUpdate",
		"PlayerPathMissionUpdate", 
    "FriendshipAdd", 
    "FriendshipPostRemove", 
    "FriendshipUpdate"
	}

	for _, str in pairs(arRewardEvents) do
		Apollo.RegisterEventHandler(str, "OnRewardsUpdate", self)
	end  
  
  -- xml
  self.xmlDoc = XmlDoc.CreateFromFile("AlpacaPlates.xml")
  
  -- register
  AlpacaPluginLibrary:RegisterPlugin(self, "Nameplates")
end

-- Events

function AlpacaPlates:OnUnitCreated(unit)
  table.insert(self.arPendingUnitIds, unit:GetId())
end

function AlpacaPlates:OnUnitDestroyed(unit)
  local fnFindIndex = function(nUnitId)
    for i, nPendingUnitId in ipairs(self.arPendingUnitIds) do
      if nPendingUnitId == nUnitId then
        return i
      end
    end
  end

  local nIndex = fnFindIndex(unit:GetId()) 
  
  if nIndex then
    table.remove(self.arPendingUnitIds, nIndex)
  end
end

function AlpacaPlates:OnUnitTextBubbleToggled(unit, strText)
	local wndNameplate = self.tActiveNameplates[unit:GetId()]
  
	if wndNameplate then
    local tData = wndNameplate:GetData()
    
    tData.bIsChatting = strText and strText ~= ""
    tData.bRefreshFull = true
	end
end

function AlpacaPlates:OnUnitNameChanged(unit, strNewName)
  local wndNameplate = self.tActiveNameplates[unit:GetId()]
  
  if wndNameplate then
    self:RefreshName(wndNameplate)
  end
end

function AlpacaPlates:OnUnitLevelChanged(unit)
  local wndNameplate = self.tActiveNameplates[unit:GetId()]
  
  if wndNameplate then
    self:RefreshLevel(wndNameplate)
  end  
end

function AlpacaPlates:OnUnitEnteredCombat(unit, bInCombat)
  local wndNameplate = self.tActiveNameplates[unit:GetId()]
  
	if wndNameplate then
    wndNameplate:GetData().bRefreshFull = true
	end
end

function AlpacaPlates:OnTargetUnitChanged(unit)
  -- ignore for now
end

function AlpacaPlates:OnRewardsUpdate(unit)
  -- ignore for now 
end

function AlpacaPlates:OnWorldLocationOnScreen(wndHandler, wndControl, bOnScreen)
  local tData = wndHandler:GetData()

  if tData then
    tData.bRefreshVisibility = true
    tData.bRefreshFull = true
    
    self:RefreshNameplate(wndHandler)
  end
end

function AlpacaPlates:OnUnitOcclusionChanged(wndHandler, wndControl, bOccluded)
  local tData = wndHandler:GetData()   
  
  if tData then
    tData.bRefreshVisibility = true
    tData.bRefreshFull = true

    self:RefreshNameplate(wndHandler)
  end
end

function AlpacaPlates:OnMouseButtonDown(wndHandler, wndControl, eMouseButton)
  if wndHandler == wndControl and eMouseButton == GameLib.CodeEnumInputMouse.Left then
    local tData = AlpacaWindowLibrary:FindParent(wndHandler, "Nameplate"):GetData()
       
    if GameLib.GetTargetUnit() ~= tData.unitOwner then
      GameLib.SetTargetUnit(tData.unitOwner)
    end
  end
end

-- Settings

function AlpacaPlates:GetUnitSettings(unit)
  local tUnitStyles = nil

  if unit == GameLib.GetTargetUnit() then
    tUnitStyles = self.tSettings.tTarget
  elseif unit == self.unitPlayer then 
    tUnitStyles = self.tSettings.tUserPlayer
  else
    
    local tLookup = {
      [AlpacaGameLibrary.EnumUnitFlags.PlayerFriendly]            = self.tSettings.tPlayerFriendly,
      [AlpacaGameLibrary.EnumUnitFlags.PlayerFriendlyPvp]         = self.tSettings.tPlayerFriendly,
      [AlpacaGameLibrary.EnumUnitFlags.PlayerHostile]             = self.tSettings.tPlayerHostile,
      [AlpacaGameLibrary.EnumUnitFlags.PlayerHostilePvp]          = self.tSettings.tPlayerHostile,
      [AlpacaGameLibrary.EnumUnitFlags.PlayerGroupMember]         = self.tSettings.tPlayerFriendly,
      [AlpacaGameLibrary.EnumUnitFlags.PlayerGroupMemberPvp]      = self.tSettings.tPlayerFriendly,
      [AlpacaGameLibrary.EnumUnitFlags.PetFriendly]               = self.tSettings.tPetFriendly,
      [AlpacaGameLibrary.EnumUnitFlags.PetFriendlyPvp]            = self.tSettings.tPetFriendly,
      [AlpacaGameLibrary.EnumUnitFlags.PetHostile]                = self.tSettings.tPetHostile,
      [AlpacaGameLibrary.EnumUnitFlags.PetHostilePvp]             = self.tSettings.tPetHostile,
      [AlpacaGameLibrary.EnumUnitFlags.NonPlayerFriendly]         = self.tSettings.tNonPlayerFriendly,
      [AlpacaGameLibrary.EnumUnitFlags.NonPlayerHostile]          = self.tSettings.tNonPlayerHostile,
      [AlpacaGameLibrary.EnumUnitFlags.NonPlayerNeutral]          = self.tSettings.tNonPlayerNeutral
    }

    tUnitStyles = tLookup[AlpacaGameLibrary:GetUnitFlags(self.unitPlayer, unit)]
  end
    
  if tUnitStyles then
    local bInCombat = AlpacaGameLibrary:IsInCombatOrDamaged(unit)
    
    if bInCombat and (tUnitStyles.nStyleInCombat or 0) ~= 0 then
      return self.tSettings.arStyles[tUnitStyles.nStyleInCombat]
    elseif (tUnitStyles.nStyle or 0) ~= 0 then
      return self.tSettings.arStyles[tUnitStyles.nStyle]
    end
  end
end

-- Timers

function AlpacaPlates:IsReady() 
  if not self.bReady then
    if not AlpacaWindowLibrary:IsLoaded() then 
      return false
    end
      
    if not self.xmlDoc then 
      return false
    end

    if not self.wndDisplay then
      return false
    end
    
    if not self.tSettings then
      return false
    end

    self.bReady = true
  end
  
  return self.bReady
end

function AlpacaPlates:IsActive()
  if self.bActive ~= self.tSettings.bEnable then
    if not self.tSettings.bEnable then 
      for nUnitId, _ in pairs(self.tActiveNameplates) do
        self:DeleteNameplate(nUnitId)
        table.insert(self.arPendingUnitIds, nUnitId)
      end
      
      self:ReduceAvailableNameplates()
      self:EnableCarbineNameplates()
    else
      self:DisableCarbineNameplates()
    end

    self.bActive = self.tSettings.bEnable
  end
 
  return self.bActive
end

function AlpacaPlates:OnFrame()   
  if not self:IsReady() then
    return
  end
  
  if not self:IsActive() then
    return
  end

  self.unitPlayer = AlpacaGameLibrary:GetPlayerUnit()

  if not self.unitPlayer then
    return
  end
  
  self:ProcessPendingUnitIds()
  
  self:RefreshActiveNameplates()
end

function AlpacaPlates:RefreshActiveNameplates()
  local fNow = os.clock()  
  local nActive = 0
  local bRefreshVisibility = fNow > (self.fLastRefreshVisibility and self.fLastRefreshVisibility + kfRefreshVisibilityDelay or 0)
  local bRefreshFull = fNow > (self.fLastRefreshFull and self.fLastRefreshFull + kfRefreshFullDelay or 0)
  local bLogStats = fNow > (self.fLastLogStats and self.fLastLogStats + kfLogStatsDelay or 0)
  
  for _, wndNameplate in pairs(self.tActiveNameplates) do
    local tData = wndNameplate:GetData()
       
    tData.bRefreshVisibility = tData.bRefreshVisibility or bRefreshVisibility
    tData.bRefreshFull = tData.bRefreshFull or bRefreshFull
    
    if self:RefreshNameplate(wndNameplate) then
      nActive = nActive + 1
    end
  end
  
  if bRefreshVisibility then
    self.fLastRefreshVisibility = fNow
  end
  
  if bRefreshFull then
    self.fLastRefreshFull = fNow
  end
  
  if bLogStats then
    self.fLastLogStats = fNow

    AlpacaPluginLibrary:SetDebugCounter("AlpacaPlates fLastRefreshVisibility", self.fLastRefreshVisibility)
    AlpacaPluginLibrary:SetDebugCounter("AlpacaPlates fLastRefreshFull", self.fLastRefreshFull)
    AlpacaPluginLibrary:SetDebugCounter("AlpacaPlates nActive", nActive)
    AlpacaPluginLibrary:SetDebugCounter("AlpacaPlates nAvailable", #self.arAvailableNameplates) 
    AlpacaPluginLibrary:SetDebugCounter("AlpacaPlates nPending", #self.arPendingUnitIds)
  end
end

function AlpacaPlates:RefreshNameplate(wndNameplate)  
  local tData = wndNameplate:GetData()
  
  self:UpdateUnitAnchor(wndNameplate)

  if not wndNameplate:GetUnit() then
    self:DeleteNameplate(tData.nUnitId)
    return false
  end
  
  local bIsVisible = false
  
  if tData.bRefreshVisibility then
    bIsVisible = self:IsVisible(wndNameplate)
    tData.bRefreshVisibility = nil
  else
    bIsVisible = wndNameplate:IsShown()
  end
  
  if not bIsVisible then
    wndNameplate:Show(false, true)
    return true
  end
  
  self:RefreshBars(wndNameplate)
  self:RefreshInterruptArmor(wndNameplate)
  
  if tData.bRefreshFull then
    self:RefreshName(wndNameplate)
    self:RefreshLevel(wndNameplate)
    self:RefreshIcon(wndNameplate)
    self:RefreshRewards(wndNameplate)   
    self:RefreshOffsets(wndNameplate)
    self:RefreshColors(wndNameplate)
    tData.bRefreshFull = nil
  end
  
  wndNameplate:Show(true, false)
  return true
end

function AlpacaPlates:UpdateUnitAnchor(wndNameplate)
  local tData = wndNameplate:GetData()
  local unitOwner = tData.unitOwner
  local unitPreviousAnchor = wndNameplate:GetUnit()
	local unitCurrentAnchor = unitOwner:IsMounted() and unitOwner:GetUnitMount() or unitOwner
  
  if not unitPreviousAnchor or unitPreviousAnchor ~= unitCurrentAnchor then
    wndNameplate:SetUnit(unitCurrentAnchor)
  end
end

function AlpacaPlates:IsVisible(wndNameplate)
  if not wndNameplate:IsOnScreen() then
    return false
  end 
  
  if wndNameplate:IsOccluded() then
    return false
  end
  
  local tData = wndNameplate:GetData()
  local tSettings = self:GetUnitSettings(tData.unitOwner)

  if not tSettings then
    return false
  end
  
  if tData.unitOwner:IsDead() then
    return false
  end
  
  if tData.bIsChatting and self.tSettings.bHideWhileChatBubbleActive then
    return false
  end
  
  if not self:IsInRange(tData.unitOwner) then
    return false
  end

  return true  
end

function AlpacaPlates:IsInRange(unit)
  local tSettings = self:GetUnitSettings(unit)
  
  if not tSettings then
    return false
  end

	local tPosPlayer = self.unitPlayer:GetPosition()
	local tPosUnit = (unit:IsMounted() and unit:GetUnitMount() or unit):GetPosition()

	if tPosUnit == nil then
		return false
	end

	local nDeltaX = tPosUnit.x - tPosPlayer.x
	local nDeltaY = tPosUnit.y - tPosPlayer.y
	local nDeltaZ = tPosUnit.z - tPosPlayer.z

	local nDistance = (nDeltaX * nDeltaX) + (nDeltaY * nDeltaY) + (nDeltaZ * nDeltaZ)

  return nDistance <= tSettings.nRange * tSettings.nRange
end

-- Nameplate Pool

function AlpacaPlates:ProcessPendingUnitIds()
  while #self.arPendingUnitIds > 0 do        
    self:CreateNameplate(table.remove(self.arPendingUnitIds))
	end
end

function AlpacaPlates:CreateNameplate(nUnitId) 
  self:DeleteNameplate(nUnitId)
    
  local unit = GameLib.GetUnitById(nUnitId)
  
  if not unit or not unit:IsValid() then
    return
  end
  
  local eUnitType = AlpacaGameLibrary:GetUnitType(self.playerUnit, unit)
  
  if not unit:ShouldShowNamePlate() or eUnitType == AlpacaGameLibrary.EnumUnitType.Simple or eUnitType == AlpacaGameLibrary.EnumUnitType.Unknown then
    return
  end
    
	if #self.arAvailableNameplates == 0 then
		self:IncreaseAvailableNameplates()
	end
	
  local wndNameplate = table.remove(self.arAvailableNameplates)
  
  if not wndNameplate then
    return
  end

  wndNameplate:SetData({ 
      nUnitId = nUnitId,
      unitOwner = unit,
      bRefreshVisibility = true,
      bRefreshFull = true,
      wndName = wndNameplate:FindChild("Name"),
      wndLevel = wndNameplate:FindChild("Level"),
      wndIcon = wndNameplate:FindChild("Icon"),
      wndInterruptArmor = wndNameplate:FindChild("InterruptArmor"),
      wndRewards = wndNameplate:FindChild("Rewards"),
      wndHealthBar = wndNameplate:FindChild("HealthBar"),
      wndShieldBar = wndNameplate:FindChild("ShieldBar"),
      wndAbsorbBar = wndNameplate:FindChild("AbsorbBar"),
      wndCastBar = wndNameplate:FindChild("CastBar"),
      wndDisableBar = wndNameplate:FindChild("DisableBar")
  })
  
  wndNameplate:SetUnit(unit)
  
  self.tActiveNameplates[nUnitId] = wndNameplate
end

function AlpacaPlates:DeleteNameplate(nUnitId)
  local wndNameplate = self.tActiveNameplates[nUnitId]
  
  if not wndNameplate then  
    return
  end
  
  wndNameplate:Show(false, true)
  
  self.tActiveNameplates[nUnitId] = nil  
  
  wndNameplate:SetUnit(nil)
  wndNameplate:SetData(nil)

  table.insert(self.arAvailableNameplates, wndNameplate)
  
  self:ReduceAvailableNameplates(knNameplatePoolSize)
end

function AlpacaPlates:IncreaseAvailableNameplates()
  for _ = 1, knNameplatePoolSize do
    local wndNameplate = Apollo.LoadForm(self.xmlDoc, "Nameplate", "InWorldHudStratum", self)
    wndNameplate:Show(false, true)
		table.insert(self.arAvailableNameplates, wndNameplate)
	end
end

function AlpacaPlates:ReduceAvailableNameplates()
  local nKeep = (self.tSettings.bEnable and knNameplatePoolSize or 0)
  
  while #self.arAvailableNameplates > nKeep do
    local wndNameplate = table.remove(self.arAvailableNameplates)    
    wndNameplate:Destroy()
  end
end

-- Refresh

function AlpacaPlates:RefreshBars(wndNameplate)  
  local tData = wndNameplate:GetData()
  
  local fHealthPercent = AlpacaGameLibrary:GetUnitHealthPercent(tData.unitOwner)
  local fShieldPercent = AlpacaGameLibrary:GetUnitShieldPercent(tData.unitOwner)	
  local fAbsorbPercent = AlpacaGameLibrary:GetUnitAbsorbPercent(tData.unitOwner)
  local fCastPercent = AlpacaGameLibrary:GetUnitCastPercent(tData.unitOwner)
  local fDisablePercent = AlpacaGameLibrary:GetUnitDisablePercent(tData.unitOwner)

  if self.tSettings.bPreview then
    fHealthPercent = 100
    fShieldPercent = 100
    fAbsorbPercent = 100
    fCastPercent = 100
    fDisablePercent = 100
  end

  tData.wndHealthBar:SetProgress(fHealthPercent)  
  tData.wndShieldBar:SetProgress(fShieldPercent)
  tData.wndAbsorbBar:SetProgress(fAbsorbPercent)
  tData.wndCastBar:SetProgress(fCastPercent)
  tData.wndDisableBar:SetProgress(fDisablePercent)
end

function AlpacaPlates:RefreshInterruptArmor(wndNameplate)
  local tData = wndNameplate:GetData()
  local nArmor = tData.unitOwner:GetInterruptArmorValue()
  local nArmorMax = tData.unitOwner:GetInterruptArmorMax()
  
  if nArmorMax == 0 and self.tSettings.bPreview then
    nArmor = 1
    nArmorMax = 2
  end
  
  if nArmorMax == 0 or nArmor == nil then
    tData.wndInterruptArmor:SetText(nil)
    tData.wndInterruptArmor:SetSprite(nil)
  elseif nArmorMax == -1 then
    tData.wndInterruptArmor:SetText(nil)
    tData.wndInterruptArmor:SetSprite("AlpacaSprites:InterruptArmorInvulnerable")
  else
    tData.wndInterruptArmor:SetText(nArmor > 0 and nArmor or "-")
    tData.wndInterruptArmor:SetSprite("AlpacaSprites:InterruptArmor")
  end 
end

function AlpacaPlates:RefreshName(wndNameplate)
  local tData = wndNameplate:GetData()
  local tSettings = self:GetUnitSettings(tData.unitOwner)

  if not tSettings then
    return
  end

  if tSettings.tName.strAlign == "Right" then
    tData.wndName:SetTextFlags("DT_LEFT", false)
    tData.wndName:SetTextFlags("DT_CENTER", false)
    tData.wndName:SetTextFlags("DT_RIGHT", true)
  elseif tSettings.tName.strAlign == "Center" then
    tData.wndName:SetTextFlags("DT_LEFT", false)
    tData.wndName:SetTextFlags("DT_CENTER", true)
    tData.wndName:SetTextFlags("DT_RIGHT", false)
  else
    tData.wndName:SetTextFlags("DT_LEFT", true)
    tData.wndName:SetTextFlags("DT_CENTER", false)
    tData.wndName:SetTextFlags("DT_RIGHT", false)    
  end
  
  local strName = tData.unitOwner:GetName()
  
  if GameLib.GetTargetUnit() == tData.unitOwner then    
    strName = "> " .. strName .. " <"
  end
  
  tData.wndName:SetText(strName)  
end

function AlpacaPlates:RefreshLevel(wndNameplate)
  local tData = wndNameplate:GetData()
  local tSettings = self:GetUnitSettings(tData.unitOwner)

  if not tSettings then
    return
  end

  if tSettings.tLevel.strAlign == "Right" then
    tData.wndLevel:SetTextFlags("DT_LEFT", false)
    tData.wndLevel:SetTextFlags("DT_CENTER", false)
    tData.wndLevel:SetTextFlags("DT_RIGHT", true)
  elseif tSettings.tLevel.strAlign == "Center" then
    tData.wndLevel:SetTextFlags("DT_LEFT", false)
    tData.wndLevel:SetTextFlags("DT_CENTER", true)
    tData.wndLevel:SetTextFlags("DT_RIGHT", false)
  else
    tData.wndLevel:SetTextFlags("DT_LEFT", true)
    tData.wndLevel:SetTextFlags("DT_CENTER", false)
    tData.wndLevel:SetTextFlags("DT_RIGHT", false)    
  end

  tData.wndLevel:SetText(tData.unitOwner:GetLevel())
end

function AlpacaPlates:RefreshIcon(wndNameplate)
  local tData = wndNameplate:GetData()
  local tSettings = self:GetUnitSettings(tData.unitOwner)
  
  if not tSettings then
    return
  end
  
  local eClass = tData.unitOwner:GetClassId()       
  local strSprite = nil
       
  if tSettings.tIcon.bShowPlayerIcons then
    if eClass and tSettings then	
      strSprite = ktLookupClassIcon[eClass]    
    end
  end
  
  if not strSprite then 
    local eRank = tData.unitOwner:GetRank()

    if eRank == Unit.CodeEnumRank.Elite then
      if tSettings.tIcon.bShowNonPlayerEliteIcons then 
        strSprite = ktLookupRankIcon[eRank]
      end      
    else
      if tSettings.tIcon.bShowNonPlayerCommonIcons then 
        strSprite = ktLookupRankIcon[eRank]
      end
    end
  end
  
  if not strSprite and self.tSettings.bPreview then
    strSprite = ktLookupRankIcon[Unit.CodeEnumRank.Superior]
  end
  
  tData.wndIcon:SetSprite(strSprite)
end

function AlpacaPlates:RefreshRewards(wndNameplate) 
  local tData = wndNameplate:GetData()
  local tSettings = self:GetUnitSettings(tData.unitOwner)
  
  if not tSettings then
    return
  end

  if not tSettings.tRewards.bEnable or not RewardIcons or not RewardIcons.GetUnitRewardIconsForm then
    tData.wndRewards:Show(false, true)
    return
  end
  
  local tFlags = {
    bVert = false,
    bHideQuests = false,
    bHideChallenges = false,
    bHideMissions = false,
    bHidePublicEvents = false,
    bHideRivals = false,
    bHideFriends = false
  }
  
  local nLeft = tSettings.tRewards.tPosition.nHorizontal
  local nTop = tSettings.tRewards.tPosition.nVertical
  local nWidth = 0
  local nHeight = 0
   
  tData.wndRewards:SetText(nil)
    
  RewardIcons.GetUnitRewardIconsForm(tData.wndRewards, tData.unitOwner, tFlags)

  nWidth = tData.wndRewards:ArrangeChildrenHorz(0) or 0
    
  if nWidth > 0 then
    nHeight = tData.wndRewards:GetChildren()[1]:GetHeight()
  end
  
	if self.tSettings.bPreview then
    nWidth = 50
    nHeight = 20
    tData.wndRewards:SetText("Rewards")
  end

  if nWidth == 0 or nHeight == 0 then
    tData.wndRewards:Show(false, true)    
    return
  end

  if tSettings.tRewards.strAlign == "Right" then
    tData.wndRewards:SetAnchorOffsets(nLeft - nWidth, nTop - math.ceil(nHeight / 2), nLeft, nTop + math.floor(nHeight / 2))
  elseif tSettings.tRewards.strAlign == "Center" then
    tData.wndRewards:SetAnchorOffsets(nLeft - math.ceil(nWidth / 2), nTop - math.ceil(nHeight / 2), nLeft + math.floor(nWidth / 2), nTop + math.floor(nHeight / 2))
  else
    tData.wndRewards:SetAnchorOffsets(nLeft, nTop - math.ceil(nHeight / 2), nLeft + nWidth, nTop + math.floor(nHeight / 2))    
  end
  
  tData.wndRewards:Show(true, true)
end

function AlpacaPlates:RefreshOffsets(wndNameplate) 
  local fnSetPosition = function(wndFrame, tFrameSettings)
    if not tFrameSettings.bEnable then
      wndFrame:Show(false, true)
      return
    end
    
    local nHorizontal = tFrameSettings.tPosition.nHorizontal
    local nVertical = tFrameSettings.tPosition.nVertical
    
    if tFrameSettings.strAlign then
      local nWidth = 300
      local nHeight = 20
      
      local nLeft = 0
      local nRight = 0
      local nTop = nVertical - math.ceil(nHeight / 2)
      local nBottom = nVertical + math.floor(nHeight / 2)
      
      if tFrameSettings.strAlign == "Right" then
        nLeft = nHorizontal - nWidth
        nRight = nHorizontal
      elseif tFrameSettings.strAlign == "Center" then
        nLeft = nHorizontal - math.ceil(nWidth / 2)
        nRight = nHorizontal + math.floor(nWidth / 2)
      else
        nLeft = nHorizontal
        nRight = nHorizontal + nWidth
      end
      
      wndFrame:SetAnchorOffsets(nLeft, nTop, nRight, nBottom)  
    else   
      local nWidth = tFrameSettings.nSize
      local nHeight = tFrameSettings.nSize
    
      if tFrameSettings.tSize then
        nWidth = tFrameSettings.tSize.nWidth
        nHeight = tFrameSettings.tSize.nHeight
      end

      wndFrame:SetAnchorOffsets(nHorizontal, nVertical, nHorizontal + (nWidth or 0), nVertical + (nHeight or 0))  
    end
    
    wndFrame:Show(true, true)
  end
  
  local tData = wndNameplate:GetData()
  local tSettings = self:GetUnitSettings(tData.unitOwner)
  
  if not tSettings then
    return
  end

  fnSetPosition(tData.wndName, tSettings.tName)
  fnSetPosition(tData.wndLevel, tSettings.tLevel)
  fnSetPosition(tData.wndIcon, tSettings.tIcon)
  fnSetPosition(tData.wndInterruptArmor, tSettings.tInterruptArmor)
  fnSetPosition(tData.wndHealthBar, tSettings.tHealthBar)
  fnSetPosition(tData.wndShieldBar, tSettings.tShieldBar)
  fnSetPosition(tData.wndAbsorbBar, tSettings.tAbsorbBar)
  fnSetPosition(tData.wndCastBar, tSettings.tCastBar)
  fnSetPosition(tData.wndDisableBar, tSettings.tDisableBar)
end

function AlpacaPlates:RefreshColors(wndNameplate)
  local tData = wndNameplate:GetData()
  local tSettings = self:GetUnitSettings(tData.unitOwner)
  
  if not tSettings then
    return
  end

  local eUnitType = AlpacaGameLibrary:GetUnitType(self.unitPlayer, tData.unitOwner)
	local eUnitFlags = AlpacaGameLibrary:GetUnitFlags(self.unitPlayer, tData.unitOwner)
	local eUnitDifficulty = AlpacaGameLibrary:GetUnitDifficulty(self.unitPlayer, tData.unitOwner)
  
  local crUnitType = self.tSettings.tColorSets.tUnitType[eUnitType]
  local crUnitFlags = self.tSettings.tColorSets.tUnitFlags[eUnitFlags]
  local crUnitDifficulty = self.tSettings.tColorSets.tUnitDifficulty[eUnitDifficulty]
  
  local tColorLookup = {
    [keColorSet.UnitType] = crUnitType,
    [keColorSet.UnitFlags] = crUnitFlags,
    [keColorSet.UnitDifficulty] = crUnitDifficulty
  }
   
  local crName = tColorLookup[tSettings.tName.eColorSet]
  local crLevel = tColorLookup[tSettings.tLevel.eColorSet]
  local crIcon = tColorLookup[tSettings.tIcon.eColorSet]
  local crInterruptArmor = tColorLookup[tSettings.tInterruptArmor.eColorSet]
  local crHealthBar = tColorLookup[tSettings.tHealthBar.eColorSet]
  local crShieldBar = tSettings.tShieldBar.crColor
  local crAbsorbBar = tSettings.tAbsorbBar.crColor
  local crCastBar = tSettings.tCastBar.crColor
  local crDisableBar = tSettings.tDisableBar.crColor

	tData.wndName:SetTextColor(AlpacaWindowLibrary:ToApolloColor(crName))
	tData.wndLevel:SetTextColor(AlpacaWindowLibrary:ToApolloColor(crLevel))
  tData.wndIcon:SetBGColor(AlpacaWindowLibrary:ToApolloColor(crIcon))
  tData.wndInterruptArmor:SetBGColor(AlpacaWindowLibrary:ToApolloColor(crInterruptArmor))
  tData.wndInterruptArmor:SetTextColor(AlpacaWindowLibrary:ToApolloColor(crInterruptArmor))
	tData.wndHealthBar:SetBarColor(AlpacaWindowLibrary:ToApolloColor(crHealthBar))
	tData.wndShieldBar:SetBarColor(AlpacaWindowLibrary:ToApolloColor(crShieldBar))
  tData.wndAbsorbBar:SetBarColor(AlpacaWindowLibrary:ToApolloColor(crAbsorbBar))
  tData.wndCastBar:SetBarColor(AlpacaWindowLibrary:ToApolloColor(crCastBar))
  tData.wndDisableBar:SetBarColor(AlpacaWindowLibrary:ToApolloColor(crDisableBar))
end

-- Carbine Addons

function AlpacaPlates:DisableCarbineNameplates()
  local objNameplates = Apollo.GetAddon("Nameplates")
    
  if objNameplates then
    if not self.fnCarbineBackup then
      self.fnCarbineBackup = objNameplates.HelperVerifyVisibilityOptions
    end
      
    objNameplates.HelperVerifyVisibilityOptions = function () return false end
  end
end

function AlpacaPlates:EnableCarbineNameplates()
  local objNameplates = Apollo.GetAddon("Nameplates")
  
  if objNameplates and self.fnCarbineBackup then
    objNameplates.HelperVerifyVisibilityOptions = self.fnCarbineBackup
  end        
end

-- Start

local AlpacaPlatesInst = AlpacaPlates:new()
AlpacaPlatesInst:Init()
